/**
 * @file encode.c
 * C2N pulse stream encoder
 * @author Marko Mkel <msmakela@nic.funet.fi>
 */

/* Copyright  2001,2002 Marko Mkel.

   This file is part of C2NLOAD, a program for loading files to
   Commodore 8-bit computers via the C2N232 device.

   C2NLOAD is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   C2NLOAD is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

#include "c2nload.h"
#include "encode.h"
#include <string.h>

/**
 * Encode a stream of synchronization pulses
 * @param count	number of pulses
 */
static void
encodeSync (unsigned count)
{
  while (count--)
    pulsewrite (Short);
}

/**
 * Encode a bit
 * @param b	the bit to be encoded
 */
static void
encodeBit (unsigned b)
{
  if (b) {
    pulsewrite (Medium);
    pulsewrite (Short);
  }
  else {
    pulsewrite (Short);
    pulsewrite (Medium);
  }
}

/**
 * Encode a byte marker
 */
static void
encodeMarker (void)
{
  pulsewrite (Long);
  pulsewrite (Medium);
}

/**
 * Encode a byte marker and a byte
 * @param c	the byte to be encoded
 */
static void
encodeByte (unsigned char c)
{
  register unsigned i, parity;

  encodeMarker ();

  for (parity = 1, i = 8; i--; c >>= 1) {
    encodeBit (c & 1);
    if (c & 1)
      parity = !parity;
  }

  encodeBit (parity);
}

/**
 * Encode both copies of a data block
 * @param buf	the data buffer
 * @param size	length of the data in bytes
 */
static void
encodeBlock (const unsigned char* buf, unsigned size)
{
  unsigned i;
  /** data checksum */
  unsigned chk = 0;

  for (i = 0x89; i > 0x80; )
    encodeByte (i--);

  for (i = 0; i < size; i++) {
    chk ^= buf[i];
    encodeByte (buf[i]);
  }

  encodeByte (chk);
  pulsewrite (Long);
  encodeSync (80);

  for (i = 9; i; )
    encodeByte (i--);

  for (i = 0; i < size; i++)
    encodeByte (buf[i]);

  encodeByte (chk);
  pulsewrite (Long);
}

/**
 * Encode both copies of a data block in Commodore 264 format
 * @param buf	the data buffer
 * @param size	length of the data in bytes
 */
static void
encode264Block (const unsigned char* buf, unsigned size)
{
  unsigned i;
  /** data checksum */
  unsigned chk = 0;

  for (i = 0x89; i > 0x80; )
    encodeByte (i--);

  for (i = 0; i < size; i++) {
    chk ^= buf[i];
    encodeByte (buf[i]);
  }

  encodeByte (chk);
  pulsewrite (Medium);
  encodeSync (437);

  for (i = 9; i; )
    encodeByte (i--);

  for (i = 0; i < size; i++)
    encodeByte (buf[i]);

  encodeByte (chk);
}

/** Pulse stream encoder
 * @param data	the 194-byte data to be encoded
 * @param begin	number of initial synchronization pulses
 * @param intra	number of intra-block synchronization pulses
 * @param plus4	flag: encode in Commodore 264 series format
 * @param name	file name to be loaded
 * @param start	start address of the actual data
 * @param len	length of the actual data
 * @param sys	sys address (0 if none, 0xffff if disabled)
 */
void
encode (const unsigned char* data,
	unsigned begin, unsigned intra, unsigned plus4,
	const char* name, unsigned start, unsigned len, unsigned sys)
{
  static unsigned char header[192];
  memcpy (header, data, 192);
  data += 192;
  if (name) {
    /* insert the file name */
    unsigned char* const filename = header + 5;
    unsigned i;
    for (i = 0; i < (plus4 ? 17 : 16) && name[i]; i++) {
      /* convert the name from ASCII to PETSCII */
      register unsigned char ch = (unsigned char) name[i];
      if ((ch >= 0x20 && ch <= 0x3f) ||
	  (ch >= 0x5b && ch <= 0x5e));
      else if (ch >= 0x41 && ch <= 0x5a)
	ch |= 0x80;
      else if (ch >= 0x61 && ch <= 0x7a)
	ch &= ~0x20;
      else
	ch = '?';
      filename[i] = ch;
    }
  }
  len = -len;
  header[plus4 ? 22 : 21] = len, header[plus4 ? 23 : 22] = len >> 8;
  if (sys == 0xffff)
    header[182] = 3; /* "RUN" but no RETURN */
  else if (sys) {
    unsigned i, j;
    memcpy (header + 183, "SYS", 3);
    for (i = 4, j = sys; j; j /= 10, i++);
    header[182 + i] = 13;
    header[182] = i;
    for (j = sys; j; j /= 10)
      header[182 + --i] = '0' + (j % 10);
  }
  header[plus4 ? 25 : 24] = start, header[plus4 ? 26 : 25] = start >> 8;
  encodeSync (begin);
  if (plus4)
    encode264Block (header, sizeof header), pulsewrite (Medium);
  else
    encodeBlock (header, sizeof header);
  encodeSync (intra);
  len = header[3] | header[4] << 8;
  len -= header[1] | header[2] << 8;
  if (plus4)
    encode264Block (data, len);
  else
    encodeBlock (data, len);
}
