/**
 * @file faucet.c
 * The main program
 * @author Marko Mkel (msmakela@nic.funet.fi)
 */

/* Copyright  2002 Marko Mkel.

   This file is part of FAUCET, a program for processing symmetric
   pulse streams on computer data tapes.

   FAUCET is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   FAUCET is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

/** Program version */
#define VERSION "1.0"
/** Release date (day.month.year) */
#define DATE "3.11.2002"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>

#if defined __unix || defined __unix__ || defined __APPLE__
# include <termios.h>
# include <unistd.h>
# include <fcntl.h>
#endif /* defined __unix || defined __unix__ || defined __APPLE__ */

/** name of the serial device */
static const char* serialdev;

#if defined __WIN32 || defined __WIN32__
# include <windows.h>
# ifndef WIN32
#  define WIN32
# endif /* WIN32 */
/** Report a Windows error for accessing the serial device
 * @param op	the failed operation
 */
static void
report_error (const char* op)
{
  char* msg;
  long err = GetLastError ();
  FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		 FORMAT_MESSAGE_FROM_SYSTEM | 
		 FORMAT_MESSAGE_IGNORE_INSERTS,
		 0,
		 err,
		 0, /* default language */
		 (LPTSTR) &msg,
		 0,
		 0);
  AnsiToOem (msg, msg);
  fputs ("-c ", stderr);
  fputs (serialdev, stderr);
  fputs (": ", stderr);
  fputs (op, stderr);
  fprintf (stderr, " failed with code %ld: ", err);
  fputs (msg, stderr);
  LocalFree (msg);
}
#endif /* __WIN32 */

#ifdef __AMIGA__
static const char version[] = "$VER: FAUCET " VERSION " (" DATE ")";
# include <proto/exec.h>
# include <devices/serial.h>
# include <devices/timer.h>
# include <dos/dos.h>
# include <clib/alib_protos.h>
# include <clib/dos_protos.h>
#endif /* __AMIGA__ */

#ifdef WIN32
/** file handle for the serial device (0=none) */
static HANDLE serialfd = 0;
/** previous terminal input/output settings for the serial line */
static DCB tio;
/** previous timeout settings for the serial line */
static COMMTIMEOUTS tio2;
/** new terminal input/output settings for the serial line */
static DCB newtio;
/** new timeout settings for the serial line */
static COMMTIMEOUTS newtio2;
#elif defined __AMIGA__
/** extended input/output request structure */
static struct IOExtSer ser;
/** timer request */
static struct timerequest* timer;
/** dummy serial file descriptor */
# define serialfd !!ser.IOSer.io_Device
#else /* !WIN32 && !AMIGA */
/** file descriptor for the serial device (0=none) */
static int serialfd = 0;
/** previous terminal input/output settings for the serial line */
static struct termios tio;
/** new terminal input/output settings for the serial line */
static struct termios newtio;
#endif /* !WIN32 && !AMIGA */

/** Cleanup function */
static void
cleanup (void)
{
#ifdef WIN32
  if (serialfd && serialfd != INVALID_HANDLE_VALUE) {
    PurgeComm (serialfd, PURGE_TXCLEAR | PURGE_RXCLEAR);
    SetCommState (serialfd, &tio);
    SetCommTimeouts (serialfd, &tio2);
    SetCommBreak (serialfd);
    Sleep (20); /* send break signal for at least 20 milliseconds */
    ClearCommBreak (serialfd);
    CloseHandle (serialfd);
    serialfd = 0;
  }
#elif defined __AMIGA__
  if (ser.IOSer.io_Device && ser.IOSer.io_Message.mn_ReplyPort) {
    ser.IOSer.io_Command = CMD_CLEAR;
    if (DoIO ((struct IORequest*) &ser))
      fprintf (stderr, "cleanup: DoIO (CMD_CLEAR) returned %d\n",
	       ser.IOSer.io_Error);
    else {
      ser.IOSer.io_Command = SDCMD_BREAK;
      if (DoIO ((struct IORequest*) &ser))
	fprintf (stderr, "cleanup: DoIO (SDCMD_BREAK) returned %d\n",
		 ser.IOSer.io_Error);
    }
  }
  if (ser.IOSer.io_Device)
    CloseDevice ((struct IORequest*) &ser);
  if (ser.IOSer.io_Message.mn_ReplyPort)
    DeletePort (ser.IOSer.io_Message.mn_ReplyPort);
  memset (&ser, 0, sizeof ser);
  if (timer) {
    if (timer->tr_node.io_Message.mn_ReplyPort)
      DeletePort (timer->tr_node.io_Message.mn_ReplyPort);
    CloseDevice ((struct IORequest*) timer);
    DeleteExtIO ((struct IORequest*) timer);
    timer = 0;
  }
#else /* !WIN32 && !AMIGA */
  if (serialfd > 0) {
    tcsetattr (serialfd, TCSADRAIN, &tio);
    tcsendbreak (serialfd, 0);
    close (serialfd);
    serialfd = 0;
  }
#endif /* !WIN32 && !AMIGA */
}

/** Signal handler
 * @param num	the signal number
 */
static void
sig (int num)
{
  cleanup ();
  signal (num, SIG_DFL);
  raise (num);
}

/** input/output buffer */
static char buf[8192];
/** number of bytes read from or written to buf */
static unsigned datain;
#if defined __AMIGA__
/** second output buffer */
static char buf2[sizeof buf];
/** pointer to the active output buffer */
static char* actbuf = buf;
/** flag: read or write pending */
static int pending = 0;

/** Flush the input/output operation on the serial line
 * @return	the WaitIO exit status
 */
static BYTE
flushout (void)
{
  const long waitsigs = SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D |
    1L << ser.IOSer.io_Message.mn_ReplyPort->mp_SigBit;
  for (;;) {
    if (CheckIO ((struct IORequest*) &ser))
      return WaitIO ((struct IORequest*) &ser);
    if (Wait (waitsigs) & (SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D)) {
      AbortIO ((struct IORequest*) &ser);
      WaitIO ((struct IORequest*) &ser);
      raise (SIGINT);
    }
  }
}
# define datalen ser.IOSer.io_Actual
#else /* !AMIGA */
/** number of bytes in buf */
static unsigned datalen;
#endif /* !AMIGA */

/** Read a pulse from the serial line
 * @return	the read pulse
 */
static unsigned char
sp_read (void)
{
  char c;
  if (!serialfd)
    return 0;
 again:
  if (datain < datalen)
    c = buf[datain++];
  else {
#ifdef WIN32
    unsigned long len;
    if (!ReadFile (serialfd, buf, sizeof buf, &len, 0) || !len) {
      cleanup ();
      return 0;
    }
    datalen = len;
    datain = 0;
#elif defined __AMIGA__
    const long waitsigs = SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D |
      1L << timer->tr_node.io_Message.mn_ReplyPort->mp_SigBit |
      1L << ser.IOSer.io_Message.mn_ReplyPort->mp_SigBit;
    if (pending)
      goto pending;
    datain = 0;
    ser.IOSer.io_Command = CMD_READ;
    ser.IOSer.io_Data = &buf;
    ser.IOSer.io_Length = sizeof buf;
    ser.IOSer.io_Actual = 0;
    ser.IOSer.io_Flags = IOF_QUICK;
    BeginIO ((struct IORequest*) &ser);
    if (ser.IOSer.io_Flags & IOF_QUICK)
      ser.IOSer.io_Flags = 0;
    else {
      pending = 1;
    pending:
      timer->tr_node.io_Command = TR_ADDREQUEST;
      timer->tr_time.tv_secs = 2;
      timer->tr_time.tv_micro = 0;
      SendIO ((struct IORequest*) timer);
      for (;;) {
	if (Wait (waitsigs) & (SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D)) {
	  pending = 0;
	  AbortIO ((struct IORequest*) timer);
	  WaitIO ((struct IORequest*) timer);
	  AbortIO ((struct IORequest*) &ser);
	  WaitIO ((struct IORequest*) &ser);
	  raise (SIGINT);
	}
	if (CheckIO ((struct IORequest*) &ser)) {
	  pending = 0;
	  AbortIO ((struct IORequest*) timer);
	  WaitIO ((struct IORequest*) timer);
	  if (WaitIO ((struct IORequest*) &ser)) {
	    fprintf (stderr, "sp_read: WaitIO (CMD_READ) returned %d\n",
		     ser.IOSer.io_Error);
	    goto clean;
	  }
	  break;
	}
	else if (CheckIO ((struct IORequest*) timer)) {
	  WaitIO ((struct IORequest*) timer);
	  if (datain < datalen)
	    break;
	  AbortIO ((struct IORequest*) &ser);
	  WaitIO ((struct IORequest*) &ser);
	  pending = 0;
	clean:
	  cleanup ();
	  return 0;
	}
      }
    }
#else /* !WIN32 && !AMIGA */
    if (!(datalen = read (serialfd, buf, 1))) {
      cleanup ();
      return 0;
    }
    datain = 0;
#endif /* !WIN32 && !AMIGA */
    goto again;
  }
  return c;
}

/** Send a bufferful of data to the serial line */
static void
sp_send (void)
{
#ifdef WIN32
  unsigned long len;
  WriteFile (serialfd, buf, datain, &len, 0);
#elif defined __AMIGA__
  if (pending)
    flushout ();
  pending = 1;
  ser.IOSer.io_Command = CMD_WRITE;
  ser.IOSer.io_Length = datain;
  ser.IOSer.io_Data = actbuf;
  SendIO ((struct IORequest*) &ser);
  actbuf = actbuf != buf ? buf : buf2;
#else /* !WIN32 && !AMIGA */
  write (serialfd, buf, datain);
#endif /* !WIN32 && !AMIGA */
  datain = 0;
}

/** Wait until the buffer has been sent
 * @return	0 if successful
 */
static int
sp_drain (void)
{
  if (datain)
    sp_send ();
#ifdef WIN32
  if (!PurgeComm (serialfd, PURGE_TXCLEAR)) {
    report_error ("PurgeComm");
    return 2;
  }
#elif defined __AMIGA__
  if (!pending)
    return 0;
  pending = 0;
  if (flushout ()) {
    fprintf (stderr, "-c %s: sp_drain: WaitIO (CMD_WRITE) returned %d\n",
	     serialdev, ser.IOSer.io_Error);
    return 2;
  }
#else /* !WIN32 && !AMIGA */
  if (tcdrain (serialfd)) {
    perror ("tcdrain");
    return 2;
  }
#endif /* !WIN32 && !AMIGA */
  return 0;
}

/** Write a pulse to the serial line
 * @param c	the pulse to be written
 */
static void
sp_write (unsigned char c)
{
  /* enqueue the pulse if it is not a pause (0)*/
  if (c && !(c >>= 1))
    c = 1; /* the level width is half the pulse width, but not zero */
  if (c) {
#if defined __AMIGA__
    actbuf[datain++] = c;
#else /* !AMIGA */
    buf[datain++] = c;
#endif /* !AMIGA */
  }

  /* send the buffer at end of file or if it became full */
  if ((!c && datain) ||
      datain == sizeof buf) {
    sp_send ();
  }

  if (!c) {
    /* pause: flush the output buffer and wait 20 milliseconds */
    sp_drain ();
#ifdef WIN32
    Sleep (20);
#elif defined __AMIGA__
    Delay (1);
#else /* !WIN32 && !AMIGA */
    usleep (20000);
#endif /* !WIN32 && !AMIGA */
  }
}

/** Write a raw byte to the serial line
 * @param c	the raw byte to be written
 */
static void
sp_wraw (unsigned char c)
{
#if defined __AMIGA__
  actbuf[datain++] = c;
#else /* !AMIGA */
  buf[datain++] = c;
#endif /* !AMIGA */
  if (datain == sizeof buf)
    sp_send ();
}

/** Open a serial line
 * @param name		the name of the device
 * @return		zero on success, nonzero on failure
 */
static int
open_serial (const char* name)
{
  serialdev = name;
# ifdef WIN32
  if ((serialfd = CreateFile (name, GENERIC_WRITE | GENERIC_READ,
			      0, 0, OPEN_EXISTING, 0, 0)) ==
      INVALID_HANDLE_VALUE) {
    report_error ("CreateFile");
    return 2;
  }

  if (!GetCommState (serialfd, &tio)) {
    report_error ("GetCommState");
  c2n232fail:
    CloseHandle (serialfd);
    serialfd = INVALID_HANDLE_VALUE;
    return 2;
  }
  else if (!GetCommTimeouts (serialfd, &tio2)) {
    report_error ("GetCommTimeouts");
    goto c2n232fail;
  }
  else {
    memcpy (&newtio, &tio, sizeof newtio);
    memcpy (&newtio2, &tio2, sizeof newtio2);
    newtio.BaudRate = CBR_38400;
    newtio.fBinary = 1;
    newtio.fParity = newtio.fOutxCtsFlow = newtio.fOutxDsrFlow = 0;
    newtio.fDtrControl = DTR_CONTROL_DISABLE;
    newtio.fDsrSensitivity = 0;
    newtio.fOutX = 1, newtio.fInX = 0;
    newtio.fErrorChar = newtio.fNull = 0;
    newtio.fRtsControl = RTS_CONTROL_DISABLE;
    newtio.fAbortOnError = 0;
    newtio.ByteSize = 8;
    newtio.Parity = NOPARITY;
    newtio.StopBits = ONESTOPBIT;
    newtio.XonChar = 021; /* DC1, ctrl-q */
    newtio.XoffChar = 023; /* DC3, ctrl-s */

    newtio2.ReadIntervalTimeout = 100; /* 1/10th second */
    newtio2.ReadTotalTimeoutConstant = 100; /* 1/10th second */
    newtio2.ReadTotalTimeoutMultiplier =
      newtio2.WriteTotalTimeoutMultiplier =
      newtio2.WriteTotalTimeoutConstant = 0;

    if (!SetCommState (serialfd, &newtio)) {
      report_error ("SetCommState");
      goto c2n232fail;
    }
    else if (!SetCommTimeouts (serialfd, &newtio2)) {
      report_error ("SetCommTimeouts");
      goto c2n232fail;
    }
    else {
      unsigned long len;
      /* send a NUL character and await a response */
      if (!WriteFile (serialfd, "", 1, &len, 0)) {
	report_error ("WriteFile");
	goto c2n232fail;
      }
      if (!ReadFile (serialfd, buf, sizeof buf, &len, 0)) {
	report_error ("ReadFile");
	goto c2n232fail;
      }
      if (!len || buf[len - 1] != '0') {
	fputs ("-c ", stderr);
	fputs (name, stderr);
	fputs (len
	       ? ": unexpected response from C2N232 device\n"
	       : ": no response from C2N232 device\n",
	       stderr);
	goto c2n232fail;
      }

      newtio2.ReadIntervalTimeout = 100; /* 1/10th second */
      newtio2.ReadTotalTimeoutConstant = 2000; /* 2 seconds */
      if (!SetCommTimeouts (serialfd, &newtio2)) {
	report_error ("SetCommTimeouts");
	goto c2n232fail;
      }
    }
  }
# elif defined __AMIGA__
  memset (&ser, 0, sizeof ser);
  if (!(ser.IOSer.io_Message.mn_ReplyPort =
	(struct MsgPort*) CreatePort (0, 0))) {
  createport:
    fputs ("CreatePort failed\n", stderr);
  c2n232fail:
    cleanup ();
    return 2;
  }
  if (OpenDevice ((const unsigned char*) name, 0,
		  (struct IORequest*) &ser, 0)) {
    fprintf (stderr, "-c %s: OpenDevice returned %d\n",
	     name, ser.IOSer.io_Error);
    goto c2n232fail;
  }
  else {
    struct MsgPort* timerport = CreatePort (0, 0);
    if (!timerport)
      goto createport;
    if (!(timer =
	  (struct timerequest*) CreateExtIO (timerport, sizeof timer))) {
      DeletePort (timerport);
      fputs ("CreateExtIO failed\n", stderr);
      goto c2n232fail;
    }
    if (OpenDevice ((const unsigned char*) "timer.device", UNIT_VBLANK,
		    (struct IORequest*) timer, 0L)) {
      fprintf (stderr, "OpenDevice for timer.device returned %d\n",
	       timer->tr_node.io_Error);
      goto c2n232fail;
    }
  }
  ser.IOSer.io_Command = SDCMD_SETPARAMS;
  ser.io_Baud = 38400;
  ser.io_ExtFlags = 0;
  ser.io_RBufLen = 2 * sizeof buf;
  ser.io_ReadLen = ser.io_WriteLen = 8;
  ser.io_StopBits = 1;
  ser.io_SerFlags = SERF_RAD_BOOGIE | SERF_XDISABLED;
  if (DoIO ((struct IORequest*) &ser)) {
    fprintf (stderr, "-c %s: DoIO (SDCMD_SETPARAMS) returned %d\n",
	     name, ser.IOSer.io_Error);
    goto c2n232fail;
  }
  /* send a NUL character and await a response */
  ser.IOSer.io_Command = CMD_WRITE;
  ser.IOSer.io_Length = 1;
  ser.IOSer.io_Data = buf;
  *buf = 0;
  SendIO ((struct IORequest*) &ser);
  if (flushout ()) {
    fprintf (stderr, "-c %s: WaitIO (CMD_WRITE) returned %d\n",
	     name, ser.IOSer.io_Error);
    goto c2n232fail;
  }
  else {
    /* await a response */
    const long waitsigs = SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D |
      1L << timer->tr_node.io_Message.mn_ReplyPort->mp_SigBit |
      1L << ser.IOSer.io_Message.mn_ReplyPort->mp_SigBit;
    ser.IOSer.io_Command = CMD_READ;
    ser.IOSer.io_Length = sizeof buf;
    ser.IOSer.io_Data = buf;
    SendIO ((struct IORequest*) &ser);
    timer->tr_node.io_Command = TR_ADDREQUEST;
    timer->tr_time.tv_secs = 0;
    timer->tr_time.tv_micro = 500000L;
    SendIO ((struct IORequest*) timer);
    for (;;) {
      if (Wait (waitsigs) & (SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D)) {
	AbortIO ((struct IORequest*) timer);
	WaitIO ((struct IORequest*) timer);
	AbortIO ((struct IORequest*) &ser);
	WaitIO ((struct IORequest*) &ser);
	raise (SIGINT);
      }
      if (CheckIO ((struct IORequest*) &ser)) {
	AbortIO ((struct IORequest*) timer);
	WaitIO ((struct IORequest*) timer);
	if (WaitIO ((struct IORequest*) &ser)) {
	  fprintf (stderr, "-c %s: WaitIO (CMD_READ) returned %d\n",
		   name, ser.IOSer.io_Error);
	  goto c2n232fail;
	}
	break;
      }
      else if (CheckIO ((struct IORequest*) timer)) {
	WaitIO ((struct IORequest*) timer);
	AbortIO ((struct IORequest*) &ser);
	WaitIO ((struct IORequest*) &ser);
	if (ser.IOSer.io_Actual > 0)
	  break;
	fputs ("-c ", stderr);
	fputs (name, stderr);
	fputs (": no response from C2N232 device\n", stderr);
	goto c2n232fail;
      }
    }
    if (buf[ser.IOSer.io_Actual - 1] != '0') {
      fputs ("-c ", stderr);
      fputs (name, stderr);
      fputs (": unexpected response from C2N232 device\n", stderr);
      goto c2n232fail;
    }
  }
# else /* !WIN32 && !AMIGA */
  if ((serialfd = open (name, O_RDWR | O_NONBLOCK | O_NOCTTY)) < 0) {
    fputs ("-c ", stderr);
    fputs (name, stderr);
    perror (": open");
    return 2;
  }
  else {
    int flags = fcntl (serialfd, F_GETFL);
    if (flags < 0) {
      fputs ("-c ", stderr);
      fputs (name, stderr);
      perror (": fcntl(GETFL)");
      close (serialfd);
      serialfd = -1;
      return 2;
    }
    else if (flags & O_NONBLOCK &&
	     fcntl (serialfd, F_SETFL, flags & ~O_NONBLOCK) < 0) {
      fputs ("-c ", stderr);
      fputs (name, stderr);
      perror (": fcntl(SETFL)");
      close (serialfd);
      serialfd = -1;
      return 2;
    }
  }

  if (tcgetattr (serialfd, &tio)) {
    fputs ("-c ", stderr);
    fputs (name, stderr);
    perror (": tcgetattr");
    close (serialfd);
    serialfd = -1;
    return 2;
  }
  else {
    memcpy (&newtio, &tio, sizeof newtio);
    newtio.c_iflag = IGNBRK | IGNPAR | IXON;
    newtio.c_oflag = 0;
    newtio.c_cflag = CS8 | CLOCAL | CREAD;
    newtio.c_lflag = 0;
    memset (newtio.c_cc, 0, sizeof newtio.c_cc);
    newtio.c_cc[VSTART] = 021; /* DC1, ctrl-q */
    newtio.c_cc[VSTOP] = 023; /* DC3, ctrl-s */
    newtio.c_cc[VMIN] = 0; /* non-blocking read */
    newtio.c_cc[VTIME] = 1; /* 1/10th second timeout */
    cfsetispeed (&newtio, B38400), cfsetospeed (&newtio, B38400);
    tcflush (serialfd, TCIOFLUSH);
    tcsendbreak (serialfd, 0);
    if (tcsetattr (serialfd, TCSANOW, &newtio)) {
      fputs ("-c ", stderr);
      fputs (name, stderr);
      perror (": tcsetattr");
    c2n232fail:
      close (serialfd);
      serialfd = -1;
      return 2;
    }
    else {
      char last = 0;
      /* send a NUL character and await a response */
      write (serialfd, "", 1);
      for (;;) {
	int len = read (serialfd, buf, sizeof buf);
	if (!len) {
	  if (last == '0')
	    break;
	  fputs ("-c ", stderr);
	  fputs (name, stderr);
	  if (last)
	    fputs (": unexpected response from C2N232 device\n", stderr);
	  else if (errno)
	    perror (": read");
	  else
	    fputs (": no response from C2N232 device\n", stderr);
	  goto c2n232fail;
	}
	last = buf[len - 1];
      }

      newtio.c_cc[VMIN] = 0; /* read at least 0 characters */
      newtio.c_cc[VTIME] = 20; /* wait 2 seconds for end of stream */
      tcsetattr (serialfd, TCSANOW, &newtio);
    }
  }
# endif /* !WIN32 && !AMIGA */
  return 0;
}

/** Construct a histogram of pulse widths
 * @param data	the pulse stream
 * @param end	end of the pulse stream
 * @param high	(output) flag: is any pulse shorter than 33 (above 3840 Hz)
 * @return	the number of pulses
 */
static unsigned long
cntpw (const unsigned char* data,
       const unsigned char* end,
       unsigned* hist,
       int* high)
{
  unsigned long num = 0;
  for (; data < end; data++) {
    if (!*data)
      break;
    num++;
    if (*data < 33)
      *high = 1;
    if (!++hist[*data])
      hist[*data]--; /* truncate the histogram on overflow */
  }
  return num;
}

/** Construct a histogram of pulse widths
 * @param f	the pulse stream
 * @param hist	(output) the histogram
 * @param high	(output) flag: is any pulse shorter than 33 (above 3840 Hz)
 * @return	the number of pulses
 */
static unsigned long
cntfpw (FILE* f,
	unsigned* hist,
	int* high)
{
  unsigned long num = 0;
  int c;
  while ((c = fgetc (f)) != EOF) {
    if (!c)
      break;
    num++;
    if (c < 33)
      *high = 1;
    if (!++hist[c])
      hist[c]--; /* truncate the histogram on overflow */
  }
  return num;
}

/** Issue a command to the C2N232 device
 * @param cmd		the command
 * @param cmdlen	length of the command
 * @return		0 on success, nonzero on failure
 */
static int
modesel (const unsigned char* cmd,
	 unsigned cmdlen)
{
  /** the response */
  unsigned char resp;
#ifdef WIN32
  unsigned long len;
  if (!PurgeComm (serialfd, PURGE_TXCLEAR)) {
    report_error ("PurgeComm");
    return 2;
  }
  if (!WriteFile (serialfd, cmd, cmdlen, &len, 0)) {
    report_error ("WriteFile");
    return 2;
  }
  if (!ReadFile (serialfd, &resp, 1, &len, 0)) {
    report_error ("ReadFile");
    return 2;
  }
#elif defined __AMIGA__
  ser.IOSer.io_Command = CMD_WRITE;
  ser.IOSer.io_Length = cmdlen;
  ser.IOSer.io_Data = &cmd;
  if (DoIO ((struct IORequest*) &ser)) {
    fprintf (stderr, "-c %s: DoIO (CMD_WRITE) returned %d\n",
	     serialdev, ser.IOSer.io_Error);
    return 2;
  }
  ser.IOSer.io_Command = CMD_READ;
  ser.IOSer.io_Length = 1;
  ser.IOSer.io_Data = &resp;
  if (DoIO ((struct IORequest*) &ser)) {
    fprintf (stderr, "-c %s: DoIO (CMD_READ) returned %d\n",
	     serialdev, ser.IOSer.io_Error);
    return 2;
  }
#else /* !WIN32 && !AMIGA */
  if (tcdrain (serialfd)) {
    perror ("tcdrain");
    return 2;
  }
  if (cmdlen != write (serialfd, cmd, cmdlen)) {
    perror ("write");
    return 2;
  }
  if (!read (serialfd, &resp, 1)) {
    if (errno)
      perror ("read");
    else
      fputs ("no response from C2N232 device\n", stderr);
    return 2;
  }
#endif /* !WIN32 && !AMIGA */

  if ((resp & ~0x30) != cmd[0]) {
    fprintf (stderr, "C2N232 responded 0x%02x; expected 0x%02x\n",
	     resp, cmd[0] | 0x30);
    return 3;
  }

#ifdef __AMIGA__
  /* enable XON/XOFF handshaking if in playback mode */
  if (cmd[0] != 7) {
    ser.IOSer.io_Command = SDCMD_SETPARAMS;
    ser.io_SerFlags = 0;
    if (DoIO ((struct IORequest*) &ser)) {
      fprintf (stderr, "-c %s: DoIO (SDCMD_SETPARAMS) returned %d\n",
	       serialdev, ser.IOSer.io_Error);
      return 2;
    }
  }
#endif /* AMIGA */

  return 0;
}

/** Wait for the C2N232 to enter idle mode
 * @return	0 if successful
 */
static int
wait_idle (void)
{
  unsigned char resp;
  int c;
  if ((c = sp_drain ()))
    return c;
#ifdef WIN32
  {
    unsigned long len;
    if (!ReadFile (serialfd, &resp, 1, &len, 0)) {
      report_error ("ReadFile");
      return 2;
    }
  }
#elif defined __AMIGA__
  ser.IOSer.io_Command = CMD_READ;
  ser.IOSer.io_Length = 1;
  ser.IOSer.io_Data = &resp;
  if (DoIO ((struct IORequest*) &ser)) {
    fprintf (stderr, "-c %s: DoIO (CMD_READ) returned %d\n",
	     serialdev, ser.IOSer.io_Error);
    return 2;
  }
#else /* !WIN32 && !AMIGA */
  if (!read (serialfd, &resp, 1)) {
    if (errno)
      perror ("wait_idle: read");
    else
      fputs ("wait_idle: no response\n", stderr);
    return 2;
  }
#endif /* !WIN32 && !AMIGA */

  if (resp != '0') {
    fprintf (stderr, "wait_idle: unexpected response 0x%02x\n", resp);
    return 1;
  }
  return 0;
}

/** Switch the C2N232 to idle mode
 * @return	0 if successful
 */
static int
idlemode (void)
{
  const unsigned char cmd = 0;
  return modesel (&cmd, 1);
}

/** The main function
 * @param argc	argument count
 * @param argv	argument vector
 * @return	0 if successful
 */
int
main (int argc, char** argv)
{
  /** mode of operation */
  enum { Play, Record } mode = Play;
  /** verbose mode? */
  int verbose = 0;
  /** explicit pauses? */
  int pause = 0;
  /** program name */
  const char* prog = *argv;

  atexit (cleanup);
  signal (SIGINT, sig);

  /* process the options */
  for (argv++; argc > 1 && **argv == '-'; argv++, argc--) {
    const char* opts = *argv;

    if (!strcmp (opts, "--")) { /* disable processing further options */
      argv++;
      argc--;
      break;
    }

    while (*++opts) /* process all flags */
      switch (*opts) {
      case 'c':
	if (argc <= 2 || serialfd)
	  goto Usage;
	if (open_serial (*++argv))
	  return 2;
	argc--;
	break;
      case 'v':
	verbose = 1;
	break;
      case 'p':
	pause = 1;
	break;
      case 'o':
	mode = Record;
	break;
      case 'h':
      default:
	goto Usage;
      }
  }

  if (!serialfd || argc < 2) {
  Usage:
    fputs ("FAUCET " VERSION
	   " - Play or record of pulse streams via the C2N232\n"
	   "Usage: ", stderr);
    fputs (prog, stderr);
    fputs (" [options] file(s)\n"
	   "Options: -c device: specify the name of the C2N232 device\n"
	   "         -o: record (SAVE) instead of playback (LOAD)\n"
	   "         -v: verbose operation\n"
	   "         -p: wait for a keypress when the stream pauses\n"
	   "         -h: request this help text\n"
	   "         --: Stop processing any further options.\n",
	   stderr);
    return 1;
  }

  /* Process the files. */

  for (; --argc; argv++) {
    /** the .TAP file header */
    static const char tapheader[20] = "C64-TAPE-RAW\1\1\0\0\0\0\0";
    /** histogram of pulse widths */
    static unsigned pwcnt[256];
    /** length of the .TAP file */
    unsigned long taplen;
    /** start of the current .TAP section */
    long tapstart;
    /** the .TAP data */
    unsigned char* tapdata = 0;
    /** data file */
    FILE* df;
    /** most recent pulse character */
    int c;
    switch (mode) {
    case Play:
      if (!(df = fopen (*argv, "rb"))) {
	fputs (*argv, stderr);
	perror (": fopen (read)");
	continue;
      }
      /* look for the .TAP file header */
      for (c = 0; c < 13; c++) {
	if (getc (df) != tapheader[c]) {
	  fputs (*argv, stderr);
	  fputs (": not a .TAP image\n", stderr);
	  fclose (df);
	  return 3;
	}
      }
      /* skip the rest of the .TAP version number */
      if (getc (df) == EOF || getc (df) == EOF || getc (df) == EOF) {
      eof:
	fputs (*argv, stderr);
	if (ferror (df))
	  perror ("getc");
	else
	  fputs (": unexpected end of file\n", stderr);
	fclose (df);
	continue;
      }
      /* read the total length of the tape image */
      if (EOF == (c = getc (df))) goto eof;
      taplen = (unsigned char) c;
      if (EOF == (c = getc (df))) goto eof;
      taplen |= ((unsigned long) (unsigned char) c) << 8;
      if (EOF == (c = getc (df))) goto eof;
      taplen |= ((unsigned long) (unsigned char) c) << 16;
      if (EOF == (c = getc (df))) goto eof;
      taplen |= ((unsigned long) (unsigned char) c) << 24;

      /* find out if the file is seekable */
      if ((tapstart = ftell (df)) < 0) {
	fputs (*argv, stderr);
	perror (": ftell");
	fprintf (stderr, "trying to read %lu bytes into memory\n", taplen);
	if (!(tapdata = malloc (taplen))) {
	  fputs ("cannot allocate enough memory\n", stderr);
	  fclose (df);
	  continue;
	}
	if (taplen != fread (tapdata, 1, taplen, df)) {
	  free (tapdata);
	  if (feof (df))
	    goto eof;
	  fputs (*argv, stderr);
	  perror ("fread");
	  continue;
	}
	tapstart = 0;
      }

      datain = 0, datalen = 0;

      for (;;) {
	int tooshort = 0;
	unsigned long length;
	memset (pwcnt, 0, sizeof pwcnt);
	length = tapdata
	  ? cntpw (tapdata + tapstart, tapdata + taplen, pwcnt, &tooshort)
	  : cntfpw (df, pwcnt, &tooshort);
	if (!length) {
	  if (tapdata ? tapstart >= taplen : feof (df))
	    break;
	}
	else if (pause && length > 5) {
	  fprintf (stderr, "press ENTER to play %lu pulses: ", length);
	  getchar ();
	}
	else if (verbose)
	  fprintf (stderr, "playing %lu pulses\n", length);
	if (tooshort) { /* quantize the pulse widths */
	  /** number of recognised pulses, plus 1 */
	  unsigned pcnt = 1;
	  /** median pulse widths (indexed 1..4) */
	  unsigned char median[5];
	  memset (median, 0, sizeof median);
	  for (c = 16; c--; ) {
	    if (pwcnt[c]) {
	      fputs (*argv, stderr);
	      fputs (": ignoring frequencies above 7680 Hz\n", stderr);
	      break;
	    }
	  }
	  for (c = 17; c < 256; c++) {
	    if (!pwcnt[c]) {
	      /*
	       * require a gap of at least 3 units
	       * between adjacent pulse widths
	       */
	      if (!pwcnt[c + 1] && !pwcnt[c + 2] &&
		  pcnt < sizeof median && median[pcnt])
		pcnt++;
	    }
	    else if (pcnt >= sizeof median) {
	      fputs (*argv, stderr);
	      fputs (": too many pulses; ignoring the longest ones\n", stderr);
	      break;
	    }
	    else if (!median[pcnt] || pwcnt[c] > pwcnt[median[pcnt]])
	      median[pcnt] = c;
	  }
	  /* transform the histogram to a quantization table */
	  for (c = pcnt; --c; ) {
	    const unsigned m = median[c];
	    unsigned low, high;
	    low = c > 1 ? ((m + median[c - 1]) >> 1) - 1 : 0;
	    high = c < pcnt - 1 ? (m + median[c + 1]) >> 1 : 256;
	    while (high-- > low)
	      pwcnt[high] = c - 1;
	  }
	  /* select packed load operation */
	  median[0] = 8;
	  for (c = 1; c < sizeof median; c++)
	    median[c] >>= 1; /* the half pulses have half widths */
	  if ((c = idlemode ()) ||
	      (c = modesel (median, 5))) {
	    free (tapdata);
	    fclose (df);
	    return c;
	  }
	  /* pack the pulses and transmit them */
	  if (tapdata) {
	    const unsigned char* s = tapdata + tapstart;
	    unsigned byte = 0, high = 0;
	    tapstart += length + 1;
	    while (length--) {
	      if (high++) {
		sp_wraw (byte |= 0x50 * pwcnt[*s++]);
		high = 0;
	      }
	      else
		byte = 0x05 * pwcnt[*s++];
	    }
	    if (high)
	      sp_wraw (byte);
	  }
	  else {
	    unsigned byte = 0, high = 0;
	    fseek (df, tapstart, SEEK_SET);
	    tapstart += length + 1;
	    while ((c = getc (df)) != EOF && length--) {
	      if (high++) {
		sp_wraw (byte |= 0x50 * pwcnt[c]);
		high = 0;
	      }
	      else
		byte = 0x05 * pwcnt[c];
	    }
	    if (high)
	      sp_wraw (byte);
	  }
	  if ((c = wait_idle ())) {
	    free (tapdata);
	    fclose (df);
	    return c;
	  }
	}
	else {
	  /* play back the pulses directly */
	  unsigned char cmd = 7;
	  if ((c = idlemode ()) ||
	      (c = modesel (&cmd, 1))) {
	    free (tapdata);
	    fclose (df);
	    return c;
	  }
	  if (tapdata) {
	    const unsigned char* s = tapdata + tapstart;
	    tapstart += length + 1;
	    while (length--)
	      sp_write (*s++);
	  }
	  else {
	    fseek (df, tapstart, SEEK_SET);
	    tapstart += length + 1;
	    while ((c = getc (df)) != EOF && length--)
	      sp_write ((unsigned char) c);
	  }
	  sp_drain ();
	  if ((c = idlemode ()))
	    return c;
	}
      }

      free (tapdata);
      fclose (df);
      break;
    case Record:
      {
	unsigned char cmd = 1;
	if ((c = modesel (&cmd, 1)))
	  return c;
      }
      if (!(df = fopen (*argv, "wb"))) {
	fputs (*argv, stderr);
	perror (": fopen (write)");
	continue;
      }
      if (sizeof tapheader != fwrite (tapheader, 1, sizeof tapheader, df)) {
	fputs (*argv, stderr);
	perror (": fwrite");
	fclose (df);
	continue;
      }
#ifdef __AMIGA__
      if (!pending)
#endif /* AMIGA */
	datain = 0, datalen = 0;

      {
	/** maximum number of successive pauses */
	const int maxpauses = 5;
	/** number of pauses that may occur in succession */
	int pauses = maxpauses;
	/** flag: has a warning been issued? */
	int warned = 0;

	for (;;) {
	  if (!(c = sp_read ())) {
	    if (pauses--)
	      continue;
	    break;
	  }
	  else {
	    if (c < 33 && !warned) {
	      fputs (*argv, stderr);
	      fputs (": frequencies above 3840 Hz may not work properly\n",
		     stderr);
	      warned = 1;
	    }
	    if (pauses != maxpauses)
	      putc (0, df), putc (0, df);
	    pauses = maxpauses;
	    putc (c, df);
	  }
	}
      }

      /* adjust the TAP header */
      {
	long len = ftell (df) - 20;
	if (fseek (df, 16, SEEK_SET)) {
	  fputs (*argv, stderr);
	  perror (": fseek");
	}
	else {
	  putc (len, df);
	  putc (len >> 8, df);
	  putc (len >> 16, df);
	  putc (len >> 24, df);
	}
      }

      fclose (df);
      break;
    }
  }

#ifdef WIN32
  PurgeComm (serialfd, PURGE_TXCLEAR | PURGE_RXCLEAR);
#elif defined __AMIGA__
  /* nothing (for now) */
#else /* !WIN32 && !AMIGA */
  if (serialfd)
    tcdrain (serialfd);
#endif /* !WIN32 && !AMIGA */

  return 0;
}
