/*
 * Copyright  1994-1996 Marko Mkel and Olaf Seibert
 * Original Linux and Commodore 64/128/Vic-20 version by Marko Mkel
 * Ported to the PET and the Amiga series by Olaf Seibert
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * This is for the Amiga 1541 1-bit cable.
 */

#define DEBUG 		0

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <exec/exec.h>
#include <resources/misc.h>
#include <hardware/cia.h>
#include <clib/exec_protos.h>
#include <clib/misc_protos.h>

#define __PRAMIEC_C__
#include "prtrans.h"

/*
 * The CIAs are volatile, otherwise the while loop waiting for ACK to
 * change is "optimized".
 */
extern volatile struct CIA ciaa;
extern volatile struct CIA ciab;

struct Library *MiscBase;
static unsigned long MiscAllocated;
static char	myname[] = __BASE_FILE__;

#define BUSY	(1 << CIAB_PRTRBUSY)	/* == 0x01 */
#define POUT	(1 << CIAB_PRTRPOUT)	/* == 0x02 */

#define DATA		ciaa.ciaprb
#define DATA_DDR	ciaa.ciaddrb
#define CTL		ciab.ciapra
#define CTL_DDR 	ciab.ciaddra

#define ATNI		0x08		/* D3 */
#define CLKO		0x10		/* D4 */
#define DATAO		0x20		/* D5 */
#define CLKI		0x40		/* D6 */
#define DATAI		0x80		/* D7 */

/*
 * The IEC cable for Emul_1541 has inverters installed on output lines
 * and ATN in. The cable I had around does not.
 * These macros make swiching easier.
 */

#if HAVE_INVERTERS

#define INVERTERO			(DATAO | CLKO)
#define COMBINE(c,d)			(((c) | (d)) ^ INVERTERO)

#define DROP_CLK_AND_DATA		(DATA = COMBINE(0, 0))
#define RAISE_CLK_AND_DATA		(DATA = COMBINE(CLKO, DATAO))
#define DROP_CLK_DATA_1			(DATA = COMBINE(0, DATAO))
#define RAISE_CLK_DATA_1		RAISE_CLK_AND_DATA
#define DROP_DATA_CLK_1			(DATA = COMBINE(CLKO, 0))
#define RAISE_DATA_CLK_1		RAISE_CLK_AND_DATA

#else

#define INVERTERO			0

#define DROP_CLK_AND_DATA		(DATA_DDR = CLKO | DATAO)
#define RAISE_CLK_AND_DATA		(DATA_DDR = 0)
#define DROP_CLK_DATA_1			(DATA_DDR = CLKO)
#define RAISE_CLK_DATA_1		RAISE_CLK_AND_DATA
#define DROP_DATA_CLK_1			(DATA_DDR = DATAO)
#define RAISE_DATA_CLK_1		RAISE_CLK_AND_DATA

#endif

#define I2O(d)		(((d) >> 2) & (CLKO | DATAO))

#define debug(x)	/* x */

int
prinit(void)
{
    if (MiscAllocated == 0) {
	atexit(prclose);

	MiscBase = OpenResource("misc.resource");
	if (MiscBase == NULL) {
	    printf("Can't open misc.resource\n");
	    return -1;
	} {
	    const char     *user;

	    user = AllocMiscResource(MR_PARALLELPORT, myname);
	    if (user != NULL) {
		printf("Printer data lines already in use by %s.\n", user);
		return -2;
	    }
	    MiscAllocated = 1 << MR_PARALLELPORT;
	}

    }
#if HAVE_INVERTERS
    DATA = 0xFF;	/* and high */
    DATA_DDR = CLKO | DATAO;	/* set clock/data lines to output */
#else
    /*
     * If we don't have inverters, then we output a 0 by really
     * outputting a 0, but we output a 1 by turning the line to
     * input. This allows the other side to pull the line down
     * itself. Remember, in the DDR 0 = input, 1 = output.
     * This in effect looks a bit like we have inverters again.. ;-)
     */
    DATA = 0;			/* remains 0 all the time */
    DATA_DDR = 0;		/* output neutral highs */
#endif

    return 0;
}

void
prclose(void)
{
    if (MiscAllocated & (1 << MR_PARALLELPORT)) {
	DATA = 0xFF;
	DATA_DDR = 0xFF;	/* set all data lines to output */
	FreeMiscResource(MR_PARALLELPORT);
    }
    MiscAllocated = 0;
    /* There is no CloseResource() function, really! */
}

void
output(unsigned char byte)
{
    psend(&byte, 1);
}

void wait_output (unsigned char byte)
{
    /* drop CLK or DATA, depending on the bit value */
    (byte & 1) ? DROP_DATA_CLK_1 : DROP_CLK_DATA_1;

    /* wait for CLK==DATA==low */
    while ((DATA & (CLKI | DATAI)) != 0)
	usleep(SLEEP_TIME);

    /* For the rest, use the normal routine */

    psend(&byte, 1);
}

unsigned
input(void)
{
    unsigned char   byte;

    preceive(&byte, 1);
    return byte;
}

unsigned
wait_input(void)
{
    unsigned char   byte;

    /* wait for CLK==low or DATA==low */
    while ((DATA & (CLKI | DATAI)) == (CLKI | DATAI))
	usleep(SLEEP_TIME);

    /* For the rest, use the normal routine */

    preceive(&byte, 1);
    return byte;
}

void
send_switch ()
{
#if DEBUG
    printf("send_switch\n");
#endif
    do {
	/* drop CLK and DATA */
	DROP_CLK_AND_DATA;
#if DEBUG
	printf("dropped CLK and DATA\n");
#endif
	usleep(SLEEP_TIME);
	/* raise CLK and DATA */
#if DEBUG
	printf("raise CLK and DATA\n");
#endif
	RAISE_CLK_AND_DATA;
    } while ((DATA & CLKI) != 0); /* repeat until CLK==low */

    /* drop CLK (DATA remains high) */
#if DEBUG
    printf("drop CLK (DATA remains high)\n");
#endif
    DROP_CLK_DATA_1;

    /* wait for DATA==low */
#if DEBUG
    printf("wait for DATA==low\n");
#endif
    while ((DATA & DATAI) != 0);

    /* raise CLK (DATA remains high) */
#if DEBUG
    printf("raise CLK (DATA remains high)\n");
#endif
    RAISE_CLK_DATA_1;

    /* wait for DATA==high */
#if DEBUG
    printf("wait for DATA==high\n");
#endif
    while ((DATA & DATAI) == 0);

#if DEBUG
    printf("send_switch done.\n");
#endif
}

void
receive_switch()
{
#if DEBUG
    printf("receive_switch, wait for CLK==DATA==low\n");
#endif
    /* wait for CLK==DATA==low */
    while ((DATA & (CLKI | DATAI)) != 0)
	usleep(SLEEP_TIME);

    do {
	/* drop CLK (DATA remains high) */
	DROP_CLK_DATA_1;
#if DEBUG
	printf("dropped CLK (DATA remains high)\n");
#endif
	usleep(SLEEP_TIME);
	/* raise CLK */
#if DEBUG
	printf("raise CLK (DATA remains high)\n");
#endif
	RAISE_CLK_DATA_1;
    } while ((DATA & (CLKI | DATAI)) != DATAI);
    /* repeat until CLK==low && DATA==high */

    /* drop DATA (CLK remains high) */
#if DEBUG
    printf("drop DATA (CLK remains high)\n");
#endif
    DROP_DATA_CLK_1;

    /* wait for CLK==high */
#if DEBUG
    printf("wait for CLK==high\n");
#endif
    while ((DATA & CLKI) == 0);

    /* raise DATA (CLK remains high) */
#if DEBUG
    printf("raise DATA (CLK remains high)\n");
#endif
    RAISE_DATA_CLK_1;
#if DEBUG
    printf("receive_switch done\n");
#endif
}

void
psend(unsigned char *buffer, unsigned length)
{
    while (length--) {
	unsigned char byte = *buffer++;
	int i;
#if DEBUG
	printf("psend: length: %d, byte: %x\n", length+1, byte);
#endif

	for (i = 8; i--; byte >>= 1) {
#if DEBUG
	    printf("i: %d, byte: %x, drop %s\n", i, byte,
	    	(byte & 1) ? "DATA" : "CLK");
#endif
	    /* drop CLK or DATA, depending on the bit value */
	    (byte & 1) ? DROP_DATA_CLK_1 : DROP_CLK_DATA_1;

	    /* wait for CLK==DATA==low */
#if DEBUG
	    printf("wait for CLK==DATA==low\n");
#endif
	    while ((DATA & (CLKI | DATAI)) != 0);

	    /* raise CLK and DATA */
#if DEBUG
	    printf("raise CLK and DATA\n");
#endif
	    RAISE_CLK_AND_DATA;

	    /* wait for CLK==DATA==high */
#if DEBUG
	    printf("wait for CLK==DATA==high\n");
#endif
	    while ((DATA & (CLKI | DATAI)) != (CLKI | DATAI));
	}
    }
}

void
preceive(unsigned char *buffer, unsigned length)
{
    while (length--) {
	int i;
	unsigned char data = 0;
#if DEBUG
	printf("preceive: length: %d\n", length + 1);
#endif

	for (i = 8; i--; ) {
	    unsigned char regval;

	    /* wait for CLK==low or DATA==low */
#if DEBUG
	    printf("i: %d, wait for CLK==low or DATA==low\n", i);
#endif
	    do {
		regval = DATA;
	    } while ((regval & (CLKI | DATAI)) == (CLKI | DATAI));

	    /* send an acknowledgement: drop the other line */
#if DEBUG
	    printf("got %d+%02x, send an acknowledgement: drop the other line\n",
	    	(regval & CLKI) != 0, data);
#endif
#if HAVE_INVERTERS
	    DATA = I2O(regval) ^ (CLKO | DATAO) ^ INVERTERO;
#else
	    DATA_DDR = I2O(regval);
#endif

	    /* read the data */
	    data >>= 1;
	    if (regval & CLKI) data |= 128;

	    /* wait for acknowledgement (CLK==high or DATA==high) */
#if DEBUG
	    printf("wait for acknowledgement (CLK==high or DATA==high)\n");
#endif
	    while ((DATA & (CLKI | DATAI)) == 0);

	    /* raise CLK and DATA */
#if DEBUG
	    printf("raise CLK and DATA\n");
#endif
	    RAISE_CLK_AND_DATA;
	}

#if DEBUG
	printf("byte received: %02x\n", data);
#endif
	*buffer++ = data;
    }
}

