/*
 * Copyright  1994-1996 Marko Mkel and Olaf Seibert
 * Original Linux and Commodore 64/128/Vic-20 version by Marko Mkel
 * Ported to the PET and the Amiga series by Olaf Seibert
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * This is for the Amiga TransNib 4/4-bit cable.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <exec/exec.h>
#include <resources/misc.h>
#include <hardware/cia.h>
#include <clib/exec_protos.h>
#include <clib/misc_protos.h>

#define __PRAMIGATN_C__
#include "prtrans.h"

/*
 * The CIAs are volatile, otherwise the while loop waiting for ACK to
 * change is "optimized".
 */
extern volatile struct CIA ciaa;
extern volatile struct CIA ciab;

struct Library *MiscBase;
static unsigned long MiscAllocated;
static char	myname[] = __BASE_FILE__;

#define BUSY	(1 << CIAB_PRTRBUSY)	/* == 0x01 */
#define POUT	(1 << CIAB_PRTRPOUT)	/* == 0x02 */

#define DATA		ciaa.ciaprb
#define DATA_DDR	ciaa.ciaddrb
#define CTL		ciab.ciapra
#define CTL_DDR 	ciab.ciaddra

int
prinit(void)
{
    if (MiscAllocated == 0) {
	atexit(prclose);

	MiscBase = OpenResource("misc.resource");
	if (MiscBase == NULL) {
	    printf("Can't open misc.resource\n");
	    return -1;
	} {
	    const char     *user;

	    user = AllocMiscResource(MR_PARALLELPORT, myname);
	    if (user != NULL) {
		printf("Printer data lines already in use by %s.\n", user);
		return -2;
	    }
	    MiscAllocated = 1 << MR_PARALLELPORT;
	}

    }
    DATA = 0xFF;	/* and high */
    DATA_DDR = 0xFF;	/* set all lines to output */

    return 0;
}

void
prclose(void)
{
    if (MiscAllocated & (1 << MR_PARALLELPORT)) {
	DATA = 0xFF;
	DATA_DDR = 0xFF;	/* set all data lines to output */
	FreeMiscResource(MR_PARALLELPORT);
    }
    MiscAllocated = 0;
    /* There is no CloseResource() function, really! */
}

void
output(unsigned char byte)
{
    psend(&byte, 1);
}


void wait_output (unsigned char byte)
{
    register unsigned char nibble;

    DATA_DDR = 0x8f;	/* set the bidirectional signals to output */
	
    nibble = byte >> 4;

    DATA = nibble;		/* set DRDY low */
    while (DATA & 0x40)		/* wait for DRCV to go low */
	usleep (SLEEP_TIME);

    /* For the rest, use the normal routine */

    psend(&byte, 1);
}

unsigned
input(void)
{
    unsigned char   byte;

    preceive(&byte, 1);
    return byte;
}

unsigned
wait_input(void)
{
    unsigned char   byte;

    while (DATA & 0x40) /* wait for DRDY to go low */
	usleep (SLEEP_TIME);

    preceive(&byte, 1);
    return byte;
}

void send_switch() {}
void receive_switch() {}

/*-
 *  TransNib is a cable that has only 6 bits connected:
 *
 *  0-3:    data i/o
 *  6:	    handshake in
 *  7:	    handshake out
 *
 *  Because the cable is straight-through, the functions of bits 6 and 7
 *  are swapped at the PET/C-64/VIC-20 side.
 *
 *  The protocol is provided for people who made a cable for
 *  TransNib V1.00 Devised by Matt Francis (m.p.francis@newcastle.ac.uk).
 *  Because the cable is a subset of the prlink cable this protocol will
 *  also work with that cable.
 *
 *  Note that this protocol is not compatible with the TransNib programs!
-*/

void
psend(unsigned char *buffer, unsigned length)
{
    register unsigned char data;

    /*
     * $40 is DRCV (input), $80 is DRDY (output) Note that the names are
     * swapped from receiving but the data direction isn't!
     */
    DATA_DDR = 0x8f;	/* set the bidirectional signals to output */

    while (length--) {
	register unsigned char nibble;

	data = *buffer++;
	nibble = data >> 4;

	DATA = nibble;		/* set DRDY low */
	while (DATA & 0x40);	/* wait for DRCV to go low */
	DATA = nibble | 0x80;	/* set DRDY high */
	while (!(DATA & 0x40)); /* wait for DRCV to go high */

	nibble = data & 0x0F;

	DATA = nibble;		/* set DRDY low */
	while (DATA & 0x40);	/* wait for DRCV to go low */
	DATA = nibble | 0x80;	/* set DRDY high */
	while (!(DATA & 0x40)); /* wait for DRCV to go high */
    }
}

void
preceive(unsigned char *buffer, unsigned length)
{
    register unsigned char data;

    /*
     * $80 is DRCV (output), $40 is DRDY (input)
     */
    DATA_DDR = 0x80;		/* set the bidirectional signals to input */

    while (length--) {
	while (DATA & 0x40);	/* wait for DRDY to go low */
	data = (DATA & 0x0f) << 4;	/* read the high nybble */
	DATA = 0x00;		/* set DRCV low */
	while (!(DATA & 0x40)); /* wait for DRDY to go high */
	DATA = 0x80;		/* set DRCV high */

	while (DATA & 0x40);	/* wait for DRDY to go low */
	data |= DATA & 0x0f;	/* read the low nybble */
	DATA = 0x00;		/* set DRCV low */
	while (!(DATA & 0x40)); /* wait for DRDY to go high */
	DATA = 0x80;		/* set DRCV high */

	*buffer++ = data;
    }
}

